@@ -10,24 +10,25 @@
 #include "Array.h"
 #include "ArrayRef.h"
 #include "BuiltinDerivatives.h"
+#ifdef __CUDACC__
+#include "BuiltinDerivativesCUDA.cuh"
+#endif
 #include "CladConfig.h"
-#include "DynamicGraph.h"
 #include "FunctionTraits.h"
 #include "Matrix.h"
 #include "NumericalDiff.h"
 #include "Tape.h"
 
-#include <assert.h>
-#include <stddef.h>
+#include <array>
+#include <cassert>
+#include <cstddef>
 #include <cstring>
+#include <iterator>
+#include <type_traits>
 
 namespace clad {
-template <typename T, typename U> struct ValueAndAdjoint {
-  T value;
-  U adjoint;
-};
 
-  /// \returns the size of a c-style string
+/// \returns the size of a c-style string
 inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
   const char* code_copy = code;
 #ifdef __CUDACC__
@@ -42,16 +43,23 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
   return count;
 }
 
-  /// Tape type used for storing values in reverse-mode AD inside loops.
-  template <typename T>
-  using tape = tape_impl<T>;
+#ifdef __CUDACC__
+#define CUDA_ARGS bool CUDAkernel, dim3 grid, dim3 block,
+#define CUDA_REST_ARGS size_t shared_mem, cudaStream_t stream,
+#else
+#define CUDA_ARGS
+#define CUDA_REST_ARGS
+#endif
 
-  /// Add value to the end of the tape, return the same value.
-  template <typename T, typename... ArgsT>
-  CUDA_HOST_DEVICE T push(tape<T>& to, ArgsT... val) {
-    to.emplace_back(std::forward<ArgsT>(val)...);
-    return to.back();
-  }
+/// Tape type used for storing values in reverse-mode AD inside loops.
+template <typename T> using tape = tape_impl<T>;
+
+/// Add value to the end of the tape, return the same value.
+template <typename T, typename... ArgsT>
+CUDA_HOST_DEVICE T push(tape<T>& to, ArgsT... val) {
+  to.emplace_back(std::forward<ArgsT>(val)...);
+  return to.back();
+}
 
   /// Add value to the end of the tape, return the same value.
   /// A specialization for clad::array_ref types to use in reverse mode.
@@ -76,23 +84,57 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
   }
 
   /// The purpose of this function is to initialize adjoints
-  /// (or all of its differentiable fields) with 0.
-  // FIXME: Add support for objects.
-  /// Initialize a non-array variable.
-  template <typename T> CUDA_HOST_DEVICE void zero_init(T& x) { new (&x) T(); }
+  /// (or all of its iteratable elements) with 0.
+  namespace zero_init_detail {
+  template <class T> struct iterator_traits : std::iterator_traits<T> {};
+  template <> struct iterator_traits<void*> {};
+  template <> struct iterator_traits<const void*> {};
 
-  /// Initialize a non-const sized array when the size is known and is equal to
-  /// N.
-  template <typename T> CUDA_HOST_DEVICE void zero_init(T* x, std::size_t N) {
-    for (std::size_t i = 0; i < N; ++i)
-    zero_init(x[i]);
+  template <class T, class It>
+  std::integral_constant<
+      bool, !std::is_same<typename std::remove_cv<T>::type,
+                          typename iterator_traits<It>::value_type>::value>
+  is_range_check(It first, It last);
+
+  template <class T>
+  decltype(is_range_check<T>(std::begin(std::declval<const T&>()),
+                             std::end(std::declval<const T&>())))
+  is_range(int);
+  template <class T> std::false_type is_range(...);
+  } // namespace zero_init_detail
+
+  template <class T>
+  struct is_range : decltype(zero_init_detail::is_range<T>(0)) {};
+
+  template <class T> void zero_init(T& t);
+
+  template <class T,
+            typename std::enable_if<!is_range<T>::value, int>::type = 0>
+  void zero_impl(volatile T& t) {
+    // Fill an array with zeros.
+    // NOLINTNEXTLINE(cppcoreguidelines-avoid-c-arrays)
+    unsigned char tmp[sizeof(T)] = {};
+    // Transfer the zeros with the magic function memcpy which can implicitly
+    // create objects in the destination region of storage immediately prior to
+    // copying the sequence of characters to the destination [27.5.1(3)].
+    // (C++ has deprecated the volatile qualifiers. However, we drop them here
+    // to make sure things still work with codebases which still have them)
+    std::memcpy(const_cast<T*>(&t), tmp, sizeof(T));
   }
 
+  template <class T, typename std::enable_if<is_range<T>::value, int>::type = 0>
+  void zero_impl(T& t) {
+    for (auto& x : t)
+      zero_init(x);
+  }
+
+  template <class T> void zero_init(T& t) { zero_impl(t); }
+
   /// Initialize a const sized array.
   // NOLINTBEGIN(cppcoreguidelines-avoid-c-arrays)
-  template <typename T, std::size_t N>
-  CUDA_HOST_DEVICE void zero_init(T (&arr)[N]) {
-    zero_init((T*)arr, N);
+  template <typename T> CUDA_HOST_DEVICE void zero_init(T* x, std::size_t N) {
+    for (std::size_t i = 0; i < N; ++i)
+      zero_init(x[i]);
   }
   // NOLINTEND(cppcoreguidelines-avoid-c-arrays)
 
@@ -117,29 +159,56 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
   template <bool EnablePadding, class... Rest, class F, class... Args,
             class... fArgTypes,
             typename std::enable_if<EnablePadding, bool>::type = true>
-  CUDA_HOST_DEVICE return_type_t<F>
+  constexpr CUDA_HOST_DEVICE return_type_t<F>
   execute_with_default_args(list<Rest...>, F f, list<fArgTypes...>,
-                            Args&&... args) {
+                            CUDA_ARGS CUDA_REST_ARGS Args&&... args) {
+#if defined(__CUDACC__) && !defined(__CUDA_ARCH__)
+    if (CUDAkernel) {
+      constexpr size_t totalArgs = sizeof...(args) + sizeof...(Rest);
+      std::array<void*, totalArgs> argPtrs = {(void*)(&args)...,
+                                              static_cast<Rest>(nullptr)...};
+
+      void* null_param = nullptr;
+      for (size_t i = sizeof...(args); i < totalArgs; ++i)
+        argPtrs[i] = &null_param;
+
+      cudaLaunchKernel((void*)f, grid, block, argPtrs.data(), shared_mem,
+                       stream);
+      return return_type_t<F>();
+    } else {
+      return f(static_cast<Args>(args)..., static_cast<Rest>(nullptr)...);
+    }
+#else
     return f(static_cast<Args>(args)..., static_cast<Rest>(nullptr)...);
+#endif
   }
 
   template <bool EnablePadding, class... Rest, class F, class... Args,
             class... fArgTypes,
             typename std::enable_if<!EnablePadding, bool>::type = true>
-  return_type_t<F> execute_with_default_args(list<Rest...>, F f,
-                                             list<fArgTypes...>,
-                                             Args&&... args) {
+  constexpr return_type_t<F>
+  execute_with_default_args(list<Rest...>, F f, list<fArgTypes...>,
+                            CUDA_ARGS CUDA_REST_ARGS Args&&... args) {
+#if defined(__CUDACC__) && !defined(__CUDA_ARCH__)
+    if (CUDAkernel) {
+      void* argPtrs[] = {(void*)&args...};
+      cudaLaunchKernel((void*)f, grid, block, argPtrs, shared_mem, stream);
+      return return_type_t<F>();
+    }
     return f(static_cast<Args>(args)...);
+#else
+    return f(static_cast<Args>(args)...);
+#endif
   }
 
   // for executing member-functions
   template <bool EnablePadding, class... Rest, class ReturnType, class C,
             class Obj, class... Args, class... fArgTypes,
             typename std::enable_if<EnablePadding, bool>::type = true>
-  CUDA_HOST_DEVICE auto
+  constexpr CUDA_HOST_DEVICE auto
   execute_with_default_args(list<Rest...>, ReturnType C::*f, Obj&& obj,
-                            list<fArgTypes...>, Args&&... args)
-      -> return_type_t<decltype(f)> {
+                            list<fArgTypes...>,
+                            Args&&... args) -> return_type_t<decltype(f)> {
     return (static_cast<Obj>(obj).*f)((fArgTypes)(args)...,
                                       static_cast<Rest>(nullptr)...);
   }
@@ -147,9 +216,10 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
   template <bool EnablePadding, class... Rest, class ReturnType, class C,
             class Obj, class... Args, class... fArgTypes,
             typename std::enable_if<!EnablePadding, bool>::type = true>
-  auto execute_with_default_args(list<Rest...>, ReturnType C::*f, Obj&& obj,
-                                 list<fArgTypes...>, Args&&... args)
-      -> return_type_t<decltype(f)> {
+  constexpr auto
+  execute_with_default_args(list<Rest...>, ReturnType C::*f, Obj&& obj,
+                            list<fArgTypes...>,
+                            Args&&... args) -> return_type_t<decltype(f)> {
     return (static_cast<Obj>(obj).*f)(static_cast<Args>(args)...);
   }
 
@@ -161,7 +231,7 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
   /// Default value of `Functor` here is temporary, and should be removed
   /// once all clad differentiation functions support differentiating functors.
   template <typename F, typename FunctorT = ExtractFunctorTraits_t<F>,
-            bool EnablePadding = false>
+            bool EnablePadding = false, bool ImmediateMode = false>
   class CladFunction {
   public:
     using CladFunctionType = F;
@@ -169,54 +239,110 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
 
   private:
     CladFunctionType m_Function;
-    char* m_Code;
+    const char* m_Code;
     FunctorType *m_Functor = nullptr;
+    bool m_CUDAkernel = false;
 
   public:
-    CUDA_HOST_DEVICE CladFunction(CladFunctionType f,
-                                  const char* code,
-                                  FunctorType* functor = nullptr)
-        : m_Functor(functor) {
-      assert(f && "Must pass a non-0 argument.");
-      if (size_t length = GetLength(code)) {
-        m_Function = f;
-        char* temp = (char*)malloc(length + 1);
-        m_Code = temp;
-        while ((*temp++ = *code++));
-      } else {
-        // clad did not place the derivative in this object. This can happen
-        // upon error of if clad was disabled. Diagnose.
-        printf("clad failed to place the generated derivative in the object\n");
-        printf("Make sure calls to clad are within a #pragma clad ON region\n");
-
-        // Invalidate the placeholders.
-        m_Function = nullptr;
-        m_Code = nullptr;
-      }
+#ifdef __cpp_concepts
+    CUDA_HOST_DEVICE CladFunction(CladFunctionType f, const char* code,
+                                  FunctorType* functor = nullptr,
+                                  bool CUDAkernel = false)
+      requires(!ImmediateMode)
+        : m_Function(f), m_Functor(functor), m_CUDAkernel(CUDAkernel) {
+#ifndef __CLAD__
+      static_assert(false, "clad doesn't appear to be loaded; make sure that "
+                           "you pass clad.so to clang.");
+#endif
+      size_t length = GetLength(code);
+      char* temp = (char*)malloc(length + 1);
+      m_Code = temp;
+      while ((*temp++ = *code++))
+        ;
     }
+
+    constexpr CUDA_HOST_DEVICE CladFunction(CladFunctionType f,
+                                            FunctorType* functor = nullptr,
+                                            bool CUDAkernel = false)
+      requires(ImmediateMode)
+        : m_Function(f), m_Code("<constexpr functions don't have support for "
+                                "printing the derivative yet>"),
+          m_Functor(functor), m_CUDAkernel(CUDAkernel) {
+#ifndef __CLAD__
+      static_assert(false, "clad doesn't appear to be loaded; make sure that "
+                           "you pass clad.so to clang.");
+#endif
+    }
+#else
+    CUDA_HOST_DEVICE CladFunction(CladFunctionType f, const char* code,
+                                  FunctorType* functor = nullptr,
+                                  bool CUDAkernel = false)
+        : m_Function(f), m_Functor(functor), m_CUDAkernel(CUDAkernel) {
+#ifndef __CLAD__
+      static_assert(false, "clad doesn't appear to be loaded; make sure that "
+                           "you pass clad.so to clang.");
+#endif
+      size_t length = GetLength(code);
+      char* temp = (char*)malloc(length + 1);
+      m_Code = temp;
+      while ((*temp++ = *code++))
+        ;
+    }
+#endif
+
     /// Constructor overload for initializing `m_Functor` when functor
     /// is passed by reference.
-    CUDA_HOST_DEVICE
-    CladFunction(CladFunctionType f, const char* code, FunctorType& functor)
+    CUDA_HOST_DEVICE CladFunction(CladFunctionType f, const char* code,
+                                  FunctorType& functor)
         : CladFunction(f, code, &functor) {};
 
+    constexpr CUDA_HOST_DEVICE CladFunction(CladFunctionType f,
+                                            FunctorType& functor)
+        : CladFunction(f, &functor) {};
+
     // Intentionally leak m_Code, otherwise we have to link against c++ runtime,
     // i.e -lstdc++.
     //~CladFunction() { /*free(m_Code);*/ }
 
-    CladFunctionType getFunctionPtr() { return m_Function; }
+    constexpr CladFunctionType getFunctionPtr() const { return m_Function; }
 
+    template <typename... Args, class FnType = CladFunctionType>
+    typename std::enable_if<!std::is_same<FnType, NoFunction*>::value,
+                            return_type_t<F>>::type constexpr CUDA_HOST_DEVICE
+    execute(Args&&... args) const {
+      if (!m_Function)
+        return static_cast<return_type_t<F>>(return_type_t<F>());
+      if (m_CUDAkernel) {
+        printf("Use execute_kernel() for global CUDA kernels\n");
+        return static_cast<return_type_t<F>>(return_type_t<F>());
+      }
+      // here static_cast is used to achieve perfect forwarding
+#ifdef __CUDACC__
+      return execute_helper(m_Function, m_CUDAkernel, dim3(0), dim3(0),
+                            static_cast<Args>(args)...);
+#else
+      return execute_helper(m_Function, static_cast<Args>(args)...);
+#endif
+    }
+
+#ifdef __CUDACC__
     template <typename... Args, class FnType = CladFunctionType>
     typename std::enable_if<!std::is_same<FnType, NoFunction*>::value,
                             return_type_t<F>>::type
-    execute(Args&&... args) CUDA_HOST_DEVICE {
+    execute_kernel(dim3 grid, dim3 block, Args&&... args) CUDA_HOST_DEVICE {
       if (!m_Function) {
         printf("CladFunction is invalid\n");
         return static_cast<return_type_t<F>>(return_type_t<F>());
       }
-      // here static_cast is used to achieve perfect forwarding
-      return execute_helper(m_Function, static_cast<Args>(args)...);
+      if (!m_CUDAkernel) {
+        printf("Use execute() for non-global CUDA kernels\n");
+        return static_cast<return_type_t<F>>(return_type_t<F>());
+      }
+
+      return execute_helper(m_Function, m_CUDAkernel, grid, block,
+                            static_cast<Args>(args)...);
     }
+#endif
 
     /// `Execute` overload to be used when derived function type cannot be
     /// deduced. One reason for this can be when user tries to differentiate
@@ -225,18 +351,18 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
     /// subsystem.
     template <typename... Args, class FnType = CladFunctionType>
     typename std::enable_if<std::is_same<FnType, NoFunction*>::value,
-                            return_type_t<F>>::type
-    execute(Args&&... args) CUDA_HOST_DEVICE {
+                            return_type_t<F>>::type constexpr CUDA_HOST_DEVICE
+    execute(Args&&... args) const {
       return static_cast<return_type_t<F>>(0);
     }
 
     /// Return the string representation for the generated derivative.
-    const char* getCode() const {
+    constexpr const char* getCode() const {
       if (m_Code)
         return m_Code;
-      else
-        return "<invalid>";
+      return "<invalid>";
     }
+
     void dump() const {
       printf("The code is: \n%s\n", getCode());
     }
@@ -261,42 +387,64 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
     private:
       /// Helper function for executing non-member derived functions.
       template <class Fn, class... Args>
-      CUDA_HOST_DEVICE return_type_t<CladFunctionType>
-      execute_helper(Fn f, Args&&... args) {
+      constexpr CUDA_HOST_DEVICE return_type_t<CladFunctionType>
+      execute_helper(Fn f, CUDA_ARGS Args&&... args) const {
         // `static_cast` is required here for perfect forwarding.
-      return execute_with_default_args<EnablePadding>(
-          DropArgs_t<sizeof...(Args), F>{}, f,
-          TakeNFirstArgs_t<sizeof...(Args), decltype(f)>{},
-          static_cast<Args>(args)...);
+#if defined(__CUDACC__)
+        if constexpr (sizeof...(Args) >= 2) {
+          auto secondArg =
+              std::get<1>(std::forward_as_tuple(std::forward<Args>(args)...));
+          if constexpr (std::is_same<std::decay_t<decltype(secondArg)>,
+                                     cudaStream_t>::value) {
+            return [&](auto shared_mem, cudaStream_t stream, auto&&... args_) {
+              return execute_with_default_args<EnablePadding>(
+                  DropArgs_t<sizeof...(Args) - 2, F>{}, f,
+                  TakeNFirstArgs_t<sizeof...(Args) - 2, decltype(f)>{},
+                  CUDAkernel, grid, block, shared_mem, stream,
+                  static_cast<decltype(args_)>(args_)...);
+            }(static_cast<Args>(args)...);
+          } else {
+            return execute_with_default_args<EnablePadding>(
+                DropArgs_t<sizeof...(Args), F>{}, f,
+                TakeNFirstArgs_t<sizeof...(Args), decltype(f)>{}, CUDAkernel,
+                grid, block, 0, nullptr, static_cast<Args>(args)...);
+          }
+        } else {
+          return execute_with_default_args<EnablePadding>(
+              DropArgs_t<sizeof...(Args), F>{}, f,
+              TakeNFirstArgs_t<sizeof...(Args), decltype(f)>{}, CUDAkernel,
+              grid, block, 0, nullptr, static_cast<Args>(args)...);
+        }
+#else
+        return execute_with_default_args<EnablePadding>(
+            DropArgs_t<sizeof...(Args), F>{}, f,
+            TakeNFirstArgs_t<sizeof...(Args), decltype(f)>{},
+            static_cast<Args>(args)...);
+#endif
       }
 
       /// Helper functions for executing member derived functions.
       /// If user have passed object explicitly, then this specialization will
       /// be used and derived function will be called through the passed object.
-      template <
-          class ReturnType,
-          class C,
-          class Obj,
-          class = typename std::enable_if<
-              std::is_same<typename std::decay<Obj>::type, C>::value>::type,
-          class... Args>
-      return_type_t<CladFunctionType>
-      execute_helper(ReturnType C::*f, Obj&& obj, Args&&... args) {
+      template <class ReturnType, class C, class Obj,
+                class = typename std::enable_if<std::is_same<
+                    typename std::decay<Obj>::type, C>::value>::type,
+                class... Args>
+      constexpr return_type_t<CladFunctionType>
+      execute_helper(ReturnType C::*f, Obj&& obj, Args&&... args) const {
         // `static_cast` is required here for perfect forwarding.
-      return execute_with_default_args<EnablePadding>(
-          DropArgs_t<sizeof...(Args), decltype(f)>{}, f, static_cast<Obj>(obj),
-          TakeNFirstArgs_t<sizeof...(Args), decltype(f)>{},
-          static_cast<Args>(args)...);
+        return execute_with_default_args<EnablePadding>(
+            DropArgs_t<sizeof...(Args), decltype(f)>{}, f,
+            static_cast<Obj>(obj),
+            TakeNFirstArgs_t<sizeof...(Args), decltype(f)>{},
+            static_cast<Args>(args)...);
       }
       /// If user have not passed object explicitly, then this specialization
       /// will be used and derived function will be called through the object
       /// saved in `CladFunction`.
       template <class ReturnType, class C, class... Args>
-      return_type_t<CladFunctionType> execute_helper(ReturnType C::*f,
-                                                     Args&&... args) {
-        assert(m_Functor &&
-               "No default object set, explicitly pass an object to "
-               "CladFunction::execute");
+      constexpr return_type_t<CladFunctionType>
+      execute_helper(ReturnType C::*f, Args&&... args) const {
         // `static_cast` is required here for perfect forwarding.
         return execute_with_default_args<EnablePadding>(
             DropArgs_t<sizeof...(Args), decltype(f)>{}, f, *m_Functor,
@@ -328,15 +476,33 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
             typename = typename std::enable_if<
                 !clad::HasOption(GetBitmaskedOpts(BitMaskedOpts...),
                                  opts::vector_mode) &&
+                !clad::HasOption(GetBitmaskedOpts(BitMaskedOpts...),
+                                 opts::immediate_mode) &&
                 !std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
   CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>> __attribute__((
       annotate("D")))
   differentiate(F fn, ArgSpec args = "",
                 DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
                 const char* code = "") {
-      assert(fn && "Must pass in a non-0 argument");
-      return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>>(derivedFn,
-                                                                    code);
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>>(derivedFn,
+                                                                  code);
+  }
+
+  template <unsigned... BitMaskedOpts, typename ArgSpec = const char*,
+            typename F,
+            typename DerivedFnType = ExtractDerivedFnTraitsForwMode_t<F>,
+            typename = typename std::enable_if<
+                !clad::HasOption(GetBitmaskedOpts(BitMaskedOpts...),
+                                 opts::vector_mode) &&
+                clad::HasOption(GetBitmaskedOpts(BitMaskedOpts...),
+                                opts::immediate_mode) &&
+                !std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
+  constexpr CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, false,
+                         true> __attribute__((annotate("D")))
+  differentiate(F fn, ArgSpec args = "",
+                DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr)) {
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, false, true>(
+        derivedFn);
   }
 
   /// Specialization for differentiating functors.
@@ -349,13 +515,13 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
                 !clad::HasOption(GetBitmaskedOpts(BitMaskedOpts...),
                                  opts::vector_mode) &&
                 std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
-  CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>> __attribute__((
-      annotate("D")))
+  constexpr CladFunction<
+      DerivedFnType, ExtractFunctorTraits_t<F>> __attribute__((annotate("D")))
   differentiate(F&& f, ArgSpec args = "",
                 DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
                 const char* code = "") {
-      return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>>(derivedFn,
-                                                                    code, f);
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>>(derivedFn,
+                                                                  code, f);
   }
 
   /// Generates function which computes derivative of `fn` argument w.r.t
@@ -372,14 +538,13 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
                 clad::HasOption(GetBitmaskedOpts(BitMaskedOpts...),
                                 opts::vector_mode) &&
                 !std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
-  CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true> __attribute__((
-      annotate("D")))
+  constexpr CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>,
+                         true> __attribute__((annotate("D")))
   differentiate(F fn, ArgSpec args = "",
                 DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
                 const char* code = "") {
-      assert(fn && "Must pass in a non-0 argument");
-      return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true>(
-          derivedFn, code);
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true>(
+        derivedFn, code);
   }
 
   /// Generates function which computes gradient of the given function wrt the
@@ -392,15 +557,31 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
   template <unsigned... BitMaskedOpts, typename ArgSpec = const char*,
             typename F, typename DerivedFnType = GradientDerivedFnTraits_t<F>,
             typename = typename std::enable_if<
+                !clad::HasOption(GetBitmaskedOpts(BitMaskedOpts...),
+                                 opts::immediate_mode) &&
                 !std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
-  CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true> __attribute__((
-      annotate("G"))) CUDA_HOST_DEVICE
+  constexpr CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>,
+                         true> __attribute__((annotate("G"))) CUDA_HOST_DEVICE
   gradient(F f, ArgSpec args = "",
            DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
-           const char* code = "") {
-      assert(f && "Must pass in a non-0 argument");
-      return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true>(
-          derivedFn /* will be replaced by gradient*/, code);
+           const char* code = "", bool CUDAkernel = false) {
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true>(
+        derivedFn /* will be replaced by gradient*/, code, nullptr, CUDAkernel);
+  }
+
+  template <unsigned... BitMaskedOpts, typename ArgSpec = const char*,
+            typename F, typename DerivedFnType = GradientDerivedFnTraits_t<F>,
+            typename = typename std::enable_if<
+                clad::HasOption(GetBitmaskedOpts(BitMaskedOpts...),
+                                opts::immediate_mode) &&
+                !std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
+  constexpr CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true,
+                         true> __attribute__((annotate("G"))) CUDA_HOST_DEVICE
+  gradient(F f, ArgSpec args = "",
+           DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
+           bool CUDAkernel = false) {
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true, true>(
+        derivedFn /* will be replaced by gradient*/, nullptr, CUDAkernel);
   }
 
   /// Specialization for differentiating functors.
@@ -410,13 +591,13 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
             typename F, typename DerivedFnType = GradientDerivedFnTraits_t<F>,
             typename = typename std::enable_if<
                 std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
-  CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true> __attribute__((
-      annotate("G"))) CUDA_HOST_DEVICE
+  constexpr CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>,
+                         true> __attribute__((annotate("G"))) CUDA_HOST_DEVICE
   gradient(F&& f, ArgSpec args = "",
            DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
            const char* code = "") {
-      return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true>(
-          derivedFn /* will be replaced by gradient*/, code, f);
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, true>(
+        derivedFn /* will be replaced by gradient*/, code, f);
   }
 
   /// Generates function which computes hessian matrix of the given function wrt
@@ -429,15 +610,30 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
   template <unsigned... BitMaskedOpts, typename ArgSpec = const char*,
             typename F, typename DerivedFnType = HessianDerivedFnTraits_t<F>,
             typename = typename std::enable_if<
+                !clad::HasOption(GetBitmaskedOpts(BitMaskedOpts...),
+                                 opts::immediate_mode) &&
                 !std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
-  CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>> __attribute__((
-      annotate("H")))
+  constexpr CladFunction<
+      DerivedFnType, ExtractFunctorTraits_t<F>> __attribute__((annotate("H")))
   hessian(F f, ArgSpec args = "",
           DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
           const char* code = "") {
-      assert(f && "Must pass in a non-0 argument");
-      return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>>(
-          derivedFn /* will be replaced by hessian*/, code);
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>>(
+        derivedFn /* will be replaced by hessian*/, code);
+  }
+
+  template <unsigned... BitMaskedOpts, typename ArgSpec = const char*,
+            typename F, typename DerivedFnType = HessianDerivedFnTraits_t<F>,
+            typename = typename std::enable_if<
+                clad::HasOption(GetBitmaskedOpts(BitMaskedOpts...),
+                                opts::immediate_mode) &&
+                !std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
+  constexpr CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, false,
+                         true> __attribute__((annotate("H")))
+  hessian(F f, ArgSpec args = "",
+          DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr)) {
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>, false, true>(
+        derivedFn /* will be replaced by hessian*/);
   }
 
   /// Specialization for differentiating functors.
@@ -447,13 +643,13 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
             typename F, typename DerivedFnType = HessianDerivedFnTraits_t<F>,
             typename = typename std::enable_if<
                 std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
-  CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>> __attribute__((
-      annotate("H")))
+  constexpr CladFunction<
+      DerivedFnType, ExtractFunctorTraits_t<F>> __attribute__((annotate("H")))
   hessian(F&& f, ArgSpec args = "",
           DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
           const char* code = "") {
-      return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>>(
-          derivedFn /* will be replaced by hessian*/, code, f);
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>>(
+        derivedFn /* will be replaced by hessian*/, code, f);
   }
 
   /// Generates function which computes jacobian matrix of the given function
@@ -467,14 +663,14 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
             typename F, typename DerivedFnType = JacobianDerivedFnTraits_t<F>,
             typename = typename std::enable_if<
                 !std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
-  CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>> __attribute__((
-      annotate("J")))
+  constexpr CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>,
+                         /*EnablePadding=*/true> __attribute__((annotate("J")))
   jacobian(F f, ArgSpec args = "",
            DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
            const char* code = "") {
-      assert(f && "Must pass in a non-0 argument");
-      return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>>(
-          derivedFn /* will be replaced by Jacobian*/, code);
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>,
+                        /*EnablePadding=*/true>(
+        derivedFn /* will be replaced by Jacobian*/, code);
   }
 
   /// Specialization for differentiating functors.
@@ -484,22 +680,22 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
             typename F, typename DerivedFnType = JacobianDerivedFnTraits_t<F>,
             typename = typename std::enable_if<
                 std::is_class<remove_reference_and_pointer_t<F>>::value>::type>
-  CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>> __attribute__((
-      annotate("J")))
+  constexpr CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>,
+                         /*EnablePadding=*/true> __attribute__((annotate("J")))
   jacobian(F&& f, ArgSpec args = "",
            DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
            const char* code = "") {
-      return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>>(
-          derivedFn /* will be replaced by Jacobian*/, code, f);
+    return CladFunction<DerivedFnType, ExtractFunctorTraits_t<F>,
+                        /*EnablePadding=*/true>(
+        derivedFn /* will be replaced by Jacobian*/, code, f);
   }
 
   template <typename ArgSpec = const char*, typename F,
             typename DerivedFnType = GradientDerivedEstFnTraits_t<F>>
-  CladFunction<DerivedFnType> __attribute__((annotate("E")))
+  constexpr CladFunction<DerivedFnType> __attribute__((annotate("E")))
   estimate_error(F f, ArgSpec args = "",
                  DerivedFnType derivedFn = static_cast<DerivedFnType>(nullptr),
                  const char* code = "") {
-    assert(f && "Must pass in a non-0 argument");
     return CladFunction<
         DerivedFnType>(derivedFn /* will be replaced by estimation code*/,
                        code);
@@ -507,7 +703,7 @@ inline CUDA_HOST_DEVICE unsigned int GetLength(const char* code) {
 
   // Gradient Structure for Reverse Mode Enzyme
   template <unsigned N> struct EnzymeGradient { double d_arr[N]; };
-}
+  } // namespace clad
 #endif // CLAD_DIFFERENTIATOR
 
 // Enable clad after the header was included.
